#include "PipelineUI.h"
#include "CodeEditorUI.h"
#include "PropertyUI.h"
#include "PinnedUI.h"
#include "PreviewUI.h"
#include "Icons.h"
#include "../Options.h"
#include "../GUIManager.h"
#include "../Objects/Names.h"
#include "../Objects/ShaderTranscompiler.h"
#include "../Objects/SystemVariableManager.h"
#include "../Objects/ThemeContainer.h"
#include "../Engine/GLUtils.h"
#include "../Engine/GeometryFactory.h"

#include <imgui/imgui.h>
#include <imgui/imgui_internal.h>
#include <algorithm>

#define HARRAYSIZE(a) (sizeof(a)/sizeof(*a))
#define PIPELINE_SHADER_PASS_INDENT 95 * Settings::Instance().DPIScale
#define PIPELINE_ITEM_INDENT 105 * Settings::Instance().DPIScale
#define BUTTON_ICON_SIZE ImVec2(20 * Settings::Instance().DPIScale, 0)

namespace ed
{
	void PipelineUI::OnEvent(const SDL_Event& e)
	{}
	void PipelineUI::Update(float delta)
	{
		std::vector<ed::PipelineItem*>& items = m_data->Pipeline.GetList();
		
		ImVec2 containerSize = ImVec2(ImGui::GetWindowContentRegionWidth(), abs(ImGui::GetWindowContentRegionMax().y - ImGui::GetWindowContentRegionMin().y));
		ImGui::BeginChild("##object_scroll_container", containerSize);

		if (items.size() == 0)
			ImGui::TextWrapped("Right click on this window or go to Create menu in the menu bar to create a shader pass.");

		ImGui::PushStyleVar(ImGuiStyleVar_FrameBorderSize, 0);
		for (int i = 0; i < items.size(); i++) {
			m_renderItemUpDown(items, i);

			if (items[i]->Type == PipelineItem::ItemType::ShaderPass) {
				
				m_addShaderPass(items[i]);
				if (m_renderItemContext(items, i)) {
					i--;
					continue;
				}

				bool isShaderPassActive = ((pipe::ShaderPass*)items[i]->Data)->Active;
				ed::pipe::ShaderPass* data = (ed::pipe::ShaderPass*)items[i]->Data;

				bool showItems = true;
				for (int j = 0; j < m_expandList.size(); j++)
					if (m_expandList[j] == data) {
						showItems = false;
						break;
					}

				if (showItems) {
					for (int j = 0; j < data->Items.size(); j++) {
						m_renderItemUpDown(data->Items, j);
						
						if (!isShaderPassActive)
							ImGui::PushStyleVar(ImGuiStyleVar_Alpha, ImGui::GetStyle().Alpha * 0.5f);
						m_addItem(data->Items[j]);
						if (!isShaderPassActive)
							ImGui::PopStyleVar();

						if (m_renderItemContext(data->Items, j)) {
							j--;
							continue;
						}
					}
				}
			} else if (items[i]->Type == PipelineItem::ItemType::ComputePass) {
				m_addComputePass(items[i]);
				if (m_renderItemContext(items, i)) {
					i--;
					continue;
				}
			} else if (items[i]->Type == PipelineItem::ItemType::AudioPass) {
				m_addAudioPass(items[i]);
				if (m_renderItemContext(items, i)) {
					i--;
					continue;
				}
			}
		}
		ImGui::PopStyleVar();

		ImGui::EndChild();

		if (!m_itemMenuOpened && ImGui::BeginPopupContextItem("##context_main_pipeline")) {
			if (ImGui::Selectable("Create Shader Pass")) m_ui->CreateNewShaderPass();
			if (ImGui::Selectable("Create Compute Pass")) m_ui->CreateNewComputePass();
			if (ImGui::Selectable("Create Audio Pass")) m_ui->CreateNewAudioPass();
			ImGui::EndPopup();
		}
		m_itemMenuOpened = false;

		// various popups
		if (m_isVarManagerOpened) {
			ImGui::OpenPopup("Variable Manager##pui_shader_variables");
			m_isVarManagerOpened = false;
		}
		if (m_isCreateViewOpened) {
			ImGui::OpenPopup("Create Item##pui_create_item");
			m_isCreateViewOpened = false;
		}
		if (m_isChangeVarsOpened) {
			ImGui::OpenPopup("Change Variables##pui_render_variables");
			m_isChangeVarsOpened = false;
		}
		if (m_isMacroManagerOpened) {
			ImGui::OpenPopup("Shader Macros##pui_shader_macros");
			m_isMacroManagerOpened = false;
		}
		if (m_isInpLayoutManagerOpened) {
			ImGui::OpenPopup("Input layout##pui_input_layout");
			m_isInpLayoutManagerOpened = false;
		}
		if (m_isResourceManagerOpened) {
			ImGui::OpenPopup("Resource manager##pui_res_manager");
			m_isResourceManagerOpened = false;
		}

		// Shader Variable manager
		ImGui::SetNextWindowSize(ImVec2(730 * Settings::Instance().DPIScale, 225 * Settings::Instance().DPIScale), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Variable Manager##pui_shader_variables")) {
			m_renderVariableManagerUI();

			if (ImGui::Button("Ok")) m_closePopup();
			ImGui::EndPopup();
		}

		// Shader Macro Manager
		ImGui::SetNextWindowSize(ImVec2(530 * Settings::Instance().DPIScale, 175 * Settings::Instance().DPIScale), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Shader Macros##pui_shader_macros")) {
			m_renderMacroManagerUI();

			if (ImGui::Button("Ok")) m_closePopup();
			ImGui::EndPopup();
		}

		// Input Layout Manager
		ImGui::SetNextWindowSize(ImVec2(630 * Settings::Instance().DPIScale, 225 * Settings::Instance().DPIScale), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Input layout##pui_input_layout")) {
			m_renderInputLayoutManagerUI();

			if (ImGui::Button("Ok")) {
				// recreate all VAOs
				pipe::ShaderPass* pass = (pipe::ShaderPass*)m_modalItem->Data;
				std::vector<PipelineItem*>& passItems = pass->Items;

				for (auto& pitem : passItems) {
					if (pitem->Type == PipelineItem::ItemType::Geometry) {
						pipe::GeometryItem* gitem = (pipe::GeometryItem*)pitem->Data;
						
						if (gitem->Type == pipe::GeometryItem::GeometryType::ScreenQuadNDC)
							continue;

						BufferObject* bobj = (BufferObject*)gitem->InstanceBuffer;
						if (bobj == nullptr)
							gl::CreateVAO(gitem->VAO, gitem->VBO, pass->InputLayout);
						else
							gl::CreateVAO(gitem->VAO, gitem->VBO, pass->InputLayout, 0, bobj->ID, m_data->Objects.ParseBufferFormat(bobj->ViewFormat));
					} else if (pitem->Type == PipelineItem::ItemType::Model) {
						pipe::Model* mitem = (pipe::Model*)pitem->Data;
						BufferObject* bobj = (BufferObject*)mitem->InstanceBuffer;
						if (bobj == nullptr) {
							for (auto& mesh : mitem->Data->Meshes)
								gl::CreateVAO(mesh.VAO, mesh.VBO, pass->InputLayout, mesh.EBO);
						}
						else {
							for (auto& mesh : mitem->Data->Meshes)
								gl::CreateVAO(mesh.VAO, mesh.VBO, pass->InputLayout, mesh.EBO, bobj->ID, m_data->Objects.ParseBufferFormat(bobj->ViewFormat));
						}
					}
				}


				m_closePopup();
			}
			ImGui::EndPopup();
		}

		// Create Item
		ImGui::SetNextWindowSize(ImVec2(430 * Settings::Instance().DPIScale, 175 * Settings::Instance().DPIScale), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Create Item##pui_create_item")) {
			m_createUI.Update(delta);

			if (ImGui::Button("Ok")) {
				if (m_createUI.Create())
					m_closePopup();
			}
			ImGui::SameLine();
			if (ImGui::Button("Cancel")) m_closePopup();
			ImGui::EndPopup();
		}

		// variables to change when rendering specific item
		ImGui::SetNextWindowSize(ImVec2(400 * Settings::Instance().DPIScale, 175 * Settings::Instance().DPIScale), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Change Variables##pui_render_variables")) {
			m_renderChangeVariablesUI();

			if (ImGui::Button("Ok")) m_closePopup();
			ImGui::EndPopup();
		}

		// 
		ImGui::SetNextWindowSize(ImVec2(600 * Settings::Instance().DPIScale, 240 * Settings::Instance().DPIScale), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Resource manager##pui_res_manager")) {
			m_renderResourceManagerUI();

			if (ImGui::Button("Ok")) m_closePopup();
			ImGui::EndPopup();
		}
	}
	
	void PipelineUI::m_renderItemUpDown(std::vector<ed::PipelineItem*>& items, int index)
	{
		ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));

		if (ImGui::Button(std::string(UI_ICON_ARROW_UP "##U" + std::string(items[index]->Name)).c_str(), BUTTON_ICON_SIZE)) {
			if (index != 0) {
				PropertyUI* props = (reinterpret_cast<PropertyUI*>(m_ui->Get(ViewID::Properties)));
				std::string oldPropertyItemName = "";
				if (props->HasItemSelected())
					oldPropertyItemName = props->CurrentItemName();

				ed::PipelineItem* temp = items[index - 1];
				items[index - 1] = items[index];
				items[index] = temp;

				if (props->HasItemSelected()) {
					if (oldPropertyItemName == items[index - 1]->Name)
						props->Open(items[index - 1]);
					if (oldPropertyItemName == items[index]->Name)
						props->Open(items[index]);
				}
			}
		}
		ImGui::SameLine(0, 0);

		if (ImGui::Button(std::string(UI_ICON_ARROW_DOWN "##D" + std::string(items[index]->Name)).c_str(), BUTTON_ICON_SIZE)) {
			if (index != items.size() - 1) {
				PropertyUI* props = (reinterpret_cast<PropertyUI*>(m_ui->Get(ViewID::Properties)));
				std::string oldPropertyItemName = "";
				if (props->HasItemSelected())
					oldPropertyItemName = props->CurrentItemName();

				ed::PipelineItem* temp = items[index + 1];
				items[index + 1] = items[index];
				items[index] = temp;

				if (props->HasItemSelected()) {
					if (oldPropertyItemName == items[index + 1]->Name)
						props->Open(items[index + 1]);
					if (oldPropertyItemName == items[index]->Name)
						props->Open(items[index]);
				}
			}
		}
		ImGui::SameLine(0, 0);

		ImGui::PopStyleColor();
	}
	bool PipelineUI::m_renderItemContext(std::vector<ed::PipelineItem*>& items, int index)
	{
		bool ret = false; // false == we didnt delete an item

		if (ImGui::BeginPopupContextItem(("##context_" + std::string(items[index]->Name)).c_str())) {
			m_itemMenuOpened = true;
			if (items[index]->Type == PipelineItem::ItemType::ShaderPass ||
				items[index]->Type == PipelineItem::ItemType::ComputePass ||
				items[index]->Type == PipelineItem::ItemType::AudioPass)
			{
				if (ImGui::Selectable("Recompile"))
					m_data->Renderer.Recompile(items[index]->Name);

				if (items[index]->Type == PipelineItem::ItemType::ShaderPass && ImGui::BeginMenu("Add"))
				{
					if (ImGui::MenuItem("Geometry")) {
						m_isCreateViewOpened = true;
						m_createUI.SetOwner(items[index]->Name);
						m_createUI.SetType(PipelineItem::ItemType::Geometry);
					}
					else if (ImGui::MenuItem("3D Model")) {
						m_isCreateViewOpened = true;
						m_createUI.SetOwner(items[index]->Name);
						m_createUI.SetType(PipelineItem::ItemType::Model);
					}
					else if (ImGui::MenuItem("Render State")) {
						m_isCreateViewOpened = true;
						m_createUI.SetOwner(items[index]->Name);
						m_createUI.SetType(PipelineItem::ItemType::RenderState);
					}

					ImGui::EndMenu();
				}

				if (ImGui::BeginMenu("Edit Code")) {
					// TODO: show "File doesnt exist - want to create it?" msg box
					if (items[index]->Type == PipelineItem::ItemType::ShaderPass) {
						pipe::ShaderPass *passData = (pipe::ShaderPass *)(items[index]->Data);

						if (ImGui::MenuItem("Vertex Shader") && m_data->Parser.FileExists(passData->VSPath)) {
							(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->OpenVS(items[index]);
						}
						else if (ImGui::MenuItem("Pixel Shader") && m_data->Parser.FileExists(passData->PSPath)) {
							(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->OpenPS(items[index]);
						}
						else if (passData->GSUsed && ImGui::MenuItem("Geometry Shader") && m_data->Parser.FileExists(passData->GSPath)) {
							(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->OpenGS(items[index]);
						}
						else if (ImGui::MenuItem("All")) {
							if (passData->GSUsed && m_data->Parser.FileExists(passData->GSPath))
								(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->OpenGS(items[index]);

							if (m_data->Parser.FileExists(passData->PSPath))
								(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->OpenPS(items[index]);

							if (m_data->Parser.FileExists(passData->VSPath))
								(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->OpenVS(items[index]);
						}
					} else if (items[index]->Type == PipelineItem::ItemType::ComputePass) {
						pipe::ComputePass *passData = (pipe::ComputePass *)(items[index]->Data);

						if (ImGui::MenuItem("Compute Shader") && m_data->Parser.FileExists(passData->Path))
							(reinterpret_cast<CodeEditorUI *>(m_ui->Get(ViewID::Code)))->OpenCS(items[index]);
					} else if (items[index]->Type == PipelineItem::ItemType::AudioPass) {
						pipe::AudioPass *passData = (pipe::AudioPass *)(items[index]->Data);

						if (ImGui::MenuItem("Audio Shader") && m_data->Parser.FileExists(passData->Path))
							(reinterpret_cast<CodeEditorUI *>(m_ui->Get(ViewID::Code)))->OpenPS(items[index]);
					}
					
					ImGui::EndMenu();
				}

				if (ImGui::MenuItem("Variables")) {
					m_isVarManagerOpened = true;
					m_modalItem = items[index];
				}

				if (items[index]->Type == PipelineItem::ItemType::ShaderPass && ImGui::MenuItem("Input layout")) {
					m_isInpLayoutManagerOpened = true;
					m_modalItem = items[index];
				}

				if (ImGui::MenuItem("Macros")) {
					m_isMacroManagerOpened = true;
					m_modalItem = items[index];
				}

				if (ImGui::MenuItem("Resources")) {
					m_isResourceManagerOpened = true;
					m_modalItem = items[index];
				}
			}
			else if (items[index]->Type == ed::PipelineItem::ItemType::Geometry || items[index]->Type == ed::PipelineItem::ItemType::Model) {
				if (ImGui::MenuItem("Change Variables")) {
					m_isChangeVarsOpened = true;
					m_modalItem = items[index];
				}
			}

			if (ImGui::Selectable("Properties"))
				(reinterpret_cast<PropertyUI*>(m_ui->Get(ViewID::Properties)))->Open(items[index]);

			if (ImGui::Selectable("Delete")) {

				(reinterpret_cast<CodeEditorUI*>(m_ui->Get(ViewID::Code)))->CloseAllFrom(items[index]);

				// check if it is opened in property viewer
				PropertyUI* props = (reinterpret_cast<PropertyUI*>(m_ui->Get(ViewID::Properties)));
				if (props->HasItemSelected() && props->CurrentItemName() == items[index]->Name)
					props->Open(nullptr);

				if ((reinterpret_cast<PreviewUI*>(m_ui->Get(ViewID::Preview)))->IsPicked(items[index]))
					(reinterpret_cast<PreviewUI*>(m_ui->Get(ViewID::Preview)))->Pick(nullptr);

				// tell pipeline to remove this item
				m_data->Messages.ClearGroup(items[index]->Name);
				m_data->Renderer.RemoveItemVariableValues(items[index]);
				m_data->Pipeline.Remove(items[index]->Name);

				ret = true;
			}

			ImGui::EndPopup();

			return ret;
		}

		return ret;
	}

	void PipelineUI::m_closePopup()
	{
		ImGui::CloseCurrentPopup();
		m_modalItem = nullptr;
	}

	void PipelineUI::m_tooltip(const std::string& text)
	{
		if (ImGui::IsItemHovered())
		{
			ImGui::BeginTooltip();
			ImGui::PushTextWrapPos(ImGui::GetFontSize() * 35.0f);
			ImGui::TextUnformatted(text.c_str());
			ImGui::PopTextWrapPos();
			ImGui::EndTooltip();
		}
	}
	void PipelineUI::m_renderVarFlags(ed::ShaderVariable* var, char flags)
	{
		ShaderVariable::ValueType type = var->GetType();
		
		bool canInvert = type >= ShaderVariable::ValueType::Float2x2 && type <= ShaderVariable::ValueType::Float4x4;
		bool canLastFrame = var->System != ed::SystemShaderVariable::Time && var->System != ed::SystemShaderVariable::IsPicked && var->System != ed::SystemShaderVariable::ViewportSize;

		bool isInvert = var->Flags & (char)ShaderVariable::Flag::Inverse;
		bool isLastFrame = var->Flags & (char)ShaderVariable::Flag::LastFrame;

		if (var->System == ed::SystemShaderVariable::None || !canInvert) {
			ImGui::PushItemFlag(ImGuiItemFlags_Disabled, true);
			ImGui::PushStyleVar(ImGuiStyleVar_Alpha, ImGui::GetStyle().Alpha * 0.5f);
		}

		if (ImGui::Checkbox(("##flaginv" + std::string(var->Name)).c_str(), &isInvert))
			m_data->Parser.ModifyProject();
		m_tooltip("Invert");
		ImGui::SameLine();

		if (!canInvert && var->System != ed::SystemShaderVariable::None) {
			ImGui::PopStyleVar();
			ImGui::PopItemFlag();
		}

		if (!canLastFrame) {
			ImGui::PushItemFlag(ImGuiItemFlags_Disabled, true);
			ImGui::PushStyleVar(ImGuiStyleVar_Alpha, ImGui::GetStyle().Alpha * 0.5f);
		}

		if (ImGui::Checkbox(("##flaglf" + std::string(var->Name)).c_str(), &isLastFrame))
			m_data->Parser.ModifyProject();
		m_tooltip("Use last frame values");

		if (var->System == ed::SystemShaderVariable::None || !canLastFrame) {
			ImGui::PopStyleVar();
			ImGui::PopItemFlag();
		}

		var->Flags = (isInvert * (char)ShaderVariable::Flag::Inverse) |
					 (isLastFrame * (char)ShaderVariable::Flag::LastFrame);
	}
	void PipelineUI::m_renderInputLayoutManagerUI()
	{
		static InputLayoutValue iValueType = InputLayoutValue::Position;
		static char semanticName[32];

		pipe::ShaderPass *itemData = (pipe::ShaderPass*)m_modalItem->Data;

		ImGui::TextWrapped("Add or remove vertex shader inputs.");

		ImGui::BeginChild("##pui_layout_table", ImVec2(0, -25));
		ImGui::Columns(4);

		ImGui::Text("Controls"); ImGui::NextColumn();
		ImGui::Text("Location"); ImGui::NextColumn();
		ImGui::Text("Value"); ImGui::NextColumn();
		ImGui::Text("HLSL semantic"); ImGui::NextColumn();

		
		// TODO: this is only a temprorary fix for non-resizable columns
		static bool isColumnWidthSet = false;
		if (!isColumnWidthSet) {
			ImGui::SetColumnWidth(0, 80 * Settings::Instance().DPIScale);
			ImGui::SetColumnWidth(1, 60 * Settings::Instance().DPIScale);
			isColumnWidthSet = true;
		}


		ImGui::Separator();

		ImGui::PushStyleColor(ImGuiCol_FrameBg, ImVec4(0, 0, 0, 0));

		int id = 0;
		std::vector<InputLayoutItem>& els = itemData->InputLayout;

		/* EXISTING INPUTS */
		for (auto& el : els) {
			ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));

			/* UP BUTTON */
			if (ImGui::Button((UI_ICON_ARROW_UP "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE) && id != 0) {
				InputLayoutItem temp = els[id - 1];
				els[id - 1] = el;
				els[id] = temp;
				
				m_data->Parser.ModifyProject();
			}
			ImGui::SameLine(0,0);
			/* DOWN BUTTON */
			if (ImGui::Button((UI_ICON_ARROW_DOWN "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE) && id != els.size() - 1) {
				InputLayoutItem temp = els[id + 1];
				els[id + 1] = el;
				els[id] = temp;
				
				m_data->Parser.ModifyProject();
			}
			ImGui::SameLine(0,0);
			/* DELETE BUTTON */
			if (ImGui::Button((UI_ICON_DELETE "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
				els.erase(els.begin() + id);
				ImGui::PopStyleColor();
				
				m_data->Parser.ModifyProject();

				continue;
			}
			
			ImGui::PopStyleColor();
			ImGui::NextColumn();

			/* TYPE */
			ImGui::Text("%d", id);
			ImGui::NextColumn();

			/* VALUE */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			const char* valueComboPreview = ATTRIBUTE_VALUE_NAMES[(int)el.Value];
			if (ImGui::BeginCombo(("##value" + std::to_string(id)).c_str(), valueComboPreview)) {
				for (int n = 0; n < HARRAYSIZE(ATTRIBUTE_VALUE_NAMES); n++) {
					bool is_selected = (n == (int)el.Value);
					if (ImGui::Selectable(ATTRIBUTE_VALUE_NAMES[n], is_selected)) {
						el.Value = (InputLayoutValue)n;
						m_data->Parser.ModifyProject();
					}
					if (is_selected)
						ImGui::SetItemDefaultFocus();
				}
				ImGui::EndCombo();
			}
			ImGui::NextColumn();

			/* SEMANTIC */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			ImGui::Text("%s (currently not supported)", el.Semantic.c_str());
			ImGui::NextColumn();

			id++;
		}

		ImGui::PopStyleColor();

		// widgets for editing "virtual" element - an element that will be added to the list later
		/* ADD BUTTON */
		ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		if (ImGui::Button(UI_ICON_ADD)) {
			els.push_back({ iValueType, std::string(semanticName) });		
			m_data->Parser.ModifyProject();
		}
		ImGui::NextColumn();
		ImGui::PopStyleColor();

		ImGui::Text("%d", els.size());
		ImGui::NextColumn();

		/* TYPE */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		if (ImGui::Combo("##inputType", reinterpret_cast<int*>(&iValueType), ATTRIBUTE_VALUE_NAMES, HARRAYSIZE(ATTRIBUTE_VALUE_NAMES))) {
			/* Nothing to do here? */
		}
		ImGui::NextColumn();

		/* NAME */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		ImGui::InputText("##inputName", const_cast<char*>(semanticName), 32);
		ImGui::NextColumn();

		//ImGui::PopItemWidth();

		ImGui::EndChild();
		ImGui::Columns(1);
	}
	void PipelineUI::m_renderVariableManagerUI()
	{
		static ed::ShaderVariable iVariable(ed::ShaderVariable::ValueType::Float1, "var", ed::SystemShaderVariable::None);
		static ed::ShaderVariable::ValueType iValueType = ed::ShaderVariable::ValueType::Float1;
		static bool scrollToBottom = false;

		void *itemData = m_modalItem->Data;
		bool isCompute = m_modalItem->Type == PipelineItem::ItemType::ComputePass;
		bool isAudio = m_modalItem->Type == PipelineItem::ItemType::AudioPass;

		ImGui::TextWrapped("Add or remove variables bound to this shader pass.");

		ImGui::BeginChild("##pui_variable_table", ImVec2(0, -25));
		ImGui::Columns(5);

		// TODO: this is only a temprorary fix for non-resizable columns
		static bool isColumnWidthSet = false;
		if (!isColumnWidthSet) {
			ImGui::SetColumnWidth(0, (6*20) * Settings::Instance().DPIScale);
			ImGui::SetColumnWidth(3, 180 * Settings::Instance().DPIScale);
			isColumnWidthSet = true;
		}

		ImGui::Text("Controls"); ImGui::NextColumn();
		ImGui::Text("Type"); ImGui::NextColumn();
		ImGui::Text("Name"); ImGui::NextColumn();
		ImGui::Text("System"); ImGui::NextColumn();
		ImGui::Text("Flags"); ImGui::NextColumn();

		ImGui::Separator();

		ImGui::PushStyleColor(ImGuiCol_FrameBg, ImVec4(0, 0, 0, 0));

		int id = 0;
		std::vector<ed::ShaderVariable *> &els = isCompute ? ((pipe::ComputePass*)itemData)->Variables.GetVariables() : (isAudio ? ((pipe::AudioPass*)itemData)->Variables.GetVariables() : ((pipe::ShaderPass*)itemData)->Variables.GetVariables());

		/* EXISTING VARIABLES */
		for (auto& el : els) {
			ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));

			/* UP BUTTON */
			if (ImGui::Button((UI_ICON_ARROW_UP "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE) && id != 0) {
				// check if any of the affected variables are pinned
				PinnedUI* pinState = ((PinnedUI*)m_ui->Get(ViewID::Pinned));
				bool containsCur = pinState->Contains(el->Name);
				bool containsDown = pinState->Contains(els[id-1]->Name);

				m_data->Parser.ModifyProject();

				// first unpin if it was pinned
				if (containsCur)
					pinState->Remove(el->Name);
				if (containsDown)
					pinState->Remove(els[id - 1]->Name);

				ed::ShaderVariable* temp = els[id - 1];
				els[id - 1] = el;
				els[id] = temp;

				// then pin again if it was previously pinned
				if (containsCur)
					pinState->Add(els[id-1]);
				if (containsDown)
					pinState->Add(els[id]);
			}
			ImGui::SameLine(0,0);
			/* DOWN BUTTON */
			if (ImGui::Button((UI_ICON_ARROW_DOWN "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE) && id != els.size() - 1) {
				// check if any of the affected variables are pinned
				PinnedUI* pinState = ((PinnedUI*)m_ui->Get(ViewID::Pinned));
				bool containsCur = pinState->Contains(el->Name);
				bool containsDown = pinState->Contains(els[id + 1]->Name);

				m_data->Parser.ModifyProject();

				// first unpin if it was pinned
				if (containsCur)
					pinState->Remove(el->Name);
				if (containsDown)
					pinState->Remove(els[id + 1]->Name);

				ed::ShaderVariable* temp = els[id + 1];
				els[id + 1] = el;
				els[id] = temp;

				// then pin again if it was previously pinned
				if (containsCur)
					pinState->Add(els[id + 1]);
				if (containsDown)
					pinState->Add(els[id]);
			}
			ImGui::SameLine(0,0);
			/* DELETE BUTTON */
			if (ImGui::Button((UI_ICON_DELETE "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
				((PinnedUI*)m_ui->Get(ViewID::Pinned))->Remove(el->Name); // unpin if pinned
				
				m_data->Parser.ModifyProject();

				if (isCompute)
					((pipe::ComputePass*)itemData)->Variables.Remove(el->Name);
				else if (isAudio)
					((pipe::AudioPass*)itemData)->Variables.Remove(el->Name);
				else
					((pipe::ShaderPass*)itemData)->Variables.Remove(el->Name);

				ImGui::PopStyleColor();
				continue;
			}
			ImGui::SameLine(0,0);
			/* EDIT & PIN BUTTONS */
			if (el->System == ed::SystemShaderVariable::None) {
				if (ImGui::Button((UI_ICON_EDIT "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
					ImGui::OpenPopup("Value Edit##pui_shader_value_edit");
					m_valueEdit.Open(el);
				}
				ImGui::SameLine(0,0);

				PinnedUI* pinState = ((PinnedUI*)m_ui->Get(ViewID::Pinned));
				if (!pinState->Contains(el->Name)) {
					if (ImGui::Button((UI_ICON_ADD "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
						pinState->Add(el);
						m_data->Parser.ModifyProject();
					}
					m_tooltip("Pin");
				} else {
					if (ImGui::Button((UI_ICON_REMOVE "##" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
						pinState->Remove(el->Name);
						m_data->Parser.ModifyProject();
					}
					m_tooltip("Unpin");
				}

				ImGui::SameLine();
			}

			ImGui::PopStyleColor();
			ImGui::NextColumn();

			/* TYPE */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			ShaderVariable::ValueType tempType = el->GetType();
			if (ImGui::Combo(("##inputType" + std::to_string(id)).c_str(), reinterpret_cast<int*>(&tempType), VARIABLE_TYPE_NAMES, HARRAYSIZE(VARIABLE_TYPE_NAMES)))
				if (tempType != el->GetType())
					el->SetType(tempType);
			ImGui::NextColumn();

			/* NAME */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			if (ImGui::InputText(("##name" + std::to_string(id)).c_str(), const_cast<char*>(el->Name), VARIABLE_NAME_LENGTH)) {
				m_data->Parser.ModifyProject();
			}
			ImGui::NextColumn();

			/* SYSTEM VALUE */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			const char* systemComboPreview = SYSTEM_VARIABLE_NAMES[(int)el->System];
			if (ImGui::BeginCombo(("##system" + std::to_string(id)).c_str(), systemComboPreview)) {
				for (int n = 0; n < HARRAYSIZE(SYSTEM_VARIABLE_NAMES); n++) {
					bool is_selected = (n == (int)el->System);
					if ((n == 0 || ed::SystemVariableManager::GetType((ed::SystemShaderVariable)n) == el->GetType())
						&& ImGui::Selectable(SYSTEM_VARIABLE_NAMES[n], is_selected)) {
							el->System = (ed::SystemShaderVariable)n;
							m_data->Parser.ModifyProject();
						}
					if (is_selected)
						ImGui::SetItemDefaultFocus();
				}
				ImGui::EndCombo();
			}
			ImGui::NextColumn();

			/* FLAGS */
			ImGui::PopStyleColor();
			m_renderVarFlags(el, el->Flags);
			ImGui::PushStyleColor(ImGuiCol_FrameBg, ImVec4(0, 0, 0, 0));
			ImGui::NextColumn();
			
			id++;
		}

		ImGui::PopStyleColor();

		// render value edit window if needed
		ImGui::SetNextWindowSize(ImVec2(450, 175), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Value Edit##pui_shader_value_edit")) {
			m_valueEdit.Update();
			
			if (ImGui::Button("Ok")) {
				m_data->Parser.ModifyProject();
				m_valueEdit.Close();
				ImGui::CloseCurrentPopup();
			}
			ImGui::EndPopup();
		}

		// widgets for editing "virtual" element - an element that will be added to the list later
		/* ADD BUTTON */
		ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		if (ImGui::Button(UI_ICON_ADD)) {
			// cant have two variables with same name
			bool exists = false;
			for (const auto& el : els)
				if (strcmp(el->Name, iVariable.Name) == 0)
					exists = true;

			// add if it doesnt exist
			if (!exists) {
				if (isCompute)
					((pipe::ComputePass *)itemData)->Variables.AddCopy(iVariable);
				else if (isAudio)
					((pipe::AudioPass *)itemData)->Variables.AddCopy(iVariable);
				else
					((pipe::ShaderPass *)itemData)->Variables.AddCopy(iVariable);

				iVariable = ShaderVariable(ShaderVariable::ValueType::Float1, "var", ed::SystemShaderVariable::None);
				iValueType = ShaderVariable::ValueType::Float1;
				scrollToBottom = true;
				
				m_data->Parser.ModifyProject();
			}
		}
		ImGui::NextColumn();
		ImGui::PopStyleColor();

		/* TYPE */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		if (ImGui::Combo("##inputType", reinterpret_cast<int*>(&iValueType), VARIABLE_TYPE_NAMES, HARRAYSIZE(VARIABLE_TYPE_NAMES))) {
			if (iValueType != iVariable.GetType()) {
				ed::ShaderVariable newVariable(iValueType);
				memcpy(newVariable.Name, iVariable.Name, strlen(iVariable.Name));
				newVariable.System = ed::SystemShaderVariable::None;
				memcpy(newVariable.Data, iVariable.Data, std::min<int>(ed::ShaderVariable::GetSize(iVariable.GetType()), ed::ShaderVariable::GetSize(newVariable.GetType())));

				free(iVariable.Data);
				iVariable = newVariable;
			}
		}
		ImGui::NextColumn();

		/* NAME */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		ImGui::InputText("##inputName", const_cast<char*>(iVariable.Name), VARIABLE_NAME_LENGTH);
		ImGui::NextColumn();

		if (scrollToBottom) {
			ImGui::SetScrollHere();
			scrollToBottom = false;
		}

		/* SYSTEM */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		const char* inputComboPreview = SYSTEM_VARIABLE_NAMES[(int)iVariable.System];
		if (ImGui::BeginCombo(("##system" + std::to_string(id)).c_str(), inputComboPreview)) {
			for (int n = 0; n < (int)SystemShaderVariable::Count; n++) {
				bool is_selected = (n == (int)iVariable.System);
				if ((n == 0 || ed::SystemVariableManager::GetType((ed::SystemShaderVariable)n) == iVariable.GetType())
					&& ImGui::Selectable(SYSTEM_VARIABLE_NAMES[n], is_selected))
						iVariable.System = (ed::SystemShaderVariable)n;
				if (is_selected)
					ImGui::SetItemDefaultFocus();
			}
			ImGui::EndCombo();
		}
		ImGui::NextColumn();

		/* FLAGS */
		m_renderVarFlags(&iVariable, iVariable.Flags);
		ImGui::NextColumn();

		//ImGui::PopItemWidth();

		ImGui::EndChild();
		ImGui::Columns(1);
	}
	void PipelineUI::m_renderChangeVariablesUI()
	{
		static bool scrollToBottom = false;
		static int shaderTypeSel = 0, shaderVarSel = 0;
		
		ImGui::TextWrapped("Add variables that you want to change when rendering this item");

		ImGui::BeginChild("##pui_cvar_table", ImVec2(0, -25));
		ImGui::Columns(2);

		ImGui::Text("Name"); ImGui::NextColumn();
		ImGui::Text("Value"); ImGui::NextColumn();

		ImGui::Separator();

		// find this item's owner
		PipelineItem* owner = nullptr;
		pipe::ShaderPass* ownerData = nullptr;
		std::vector<PipelineItem*>& passes = m_data->Pipeline.GetList();
		for (PipelineItem* pass : passes) {
			if (pass->Type != PipelineItem::ItemType::ShaderPass)
				continue;
			
			ownerData = ((pipe::ShaderPass*)pass->Data);
			std::vector<PipelineItem*>& children = ownerData->Items;
			for (PipelineItem* child : children) {
				if (child == m_modalItem) {
					owner = pass;
					break;
				}
			}
			if (owner != nullptr)
				break;
		}

		std::vector<ShaderVariable*>& vars = ownerData->Variables.GetVariables();
		ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));

		// render the list
		int id = 0;
		std::vector<RenderEngine::ItemVariableValue> allItems = m_data->Renderer.GetItemVariableValues();
		for (auto& i : allItems) {
			if (i.Item != m_modalItem)
				continue;

			/* NAME */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			ImGui::Text(i.Variable->Name);
			ImGui::NextColumn();


			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			if (ImGui::Button((UI_ICON_EDIT "##edBtn" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
				ImGui::OpenPopup("Value Edit##pui_shader_ivalue_edit");
				m_valueEdit.Open(i.NewValue);
			}
			ImGui::SameLine();
			if (ImGui::Button((UI_ICON_DELETE "##delBtn" + std::to_string(id)).c_str(), BUTTON_ICON_SIZE)) {
				m_data->Renderer.RemoveItemVariableValue(i.Item, i.Variable);
				m_data->Parser.ModifyProject();
			}
			ImGui::NextColumn();

			id++;
		}

		ImGui::PopStyleColor();

		// render value edit window if needed
		ImGui::SetNextWindowSize(ImVec2(450, 175), ImGuiCond_Once);
		if (ImGui::BeginPopupModal("Value Edit##pui_shader_ivalue_edit")) {
			m_valueEdit.Update();

			if (ImGui::Button("Ok")) {
				m_data->Parser.ModifyProject();
				m_valueEdit.Close();
				ImGui::CloseCurrentPopup();
			}
			ImGui::EndPopup();
		}

		if (scrollToBottom) {
			ImGui::SetScrollHere();
			scrollToBottom = false;
		}

		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		const char* inputComboPreview = vars.size() > 0 ? vars[shaderVarSel]->Name : "-- NONE --";
		if (ImGui::BeginCombo(("##itemvarname" + std::to_string(id)).c_str(), inputComboPreview)) {
			for (int n = 0; n < vars.size(); n++) {
				bool is_selected = (n == shaderVarSel);
				if (ImGui::Selectable(vars[n]->Name, is_selected))
					shaderVarSel = n;
				if (is_selected)
					ImGui::SetItemDefaultFocus();
			}
			ImGui::EndCombo();
		}
		ImGui::NextColumn();

		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		if (ImGui::Button(UI_ICON_ADD, BUTTON_ICON_SIZE) && vars.size() > 0) {
			bool alreadyAdded = false;
			for (int i = 0; i < allItems.size(); i++)
				if (strcmp(vars[shaderVarSel]->Name, allItems[i].Variable->Name) == 0 && allItems[i].Item == m_modalItem) {
					alreadyAdded = true;
					break;
				}

			if (!alreadyAdded) {
				RenderEngine::ItemVariableValue itemVal(vars[shaderVarSel]);
				itemVal.Item = m_modalItem;

				m_data->Renderer.AddItemVariableValue(itemVal);
				m_data->Parser.ModifyProject();
			}
		}
		ImGui::NextColumn();

		ImGui::EndChild();
		ImGui::Columns(1);
	}
	void PipelineUI::m_renderResourceManagerUI()
	{
		ImGui::TextWrapped("List of all bound textures, images, etc...\n");
		
		ObjectManager* objs = &m_data->Objects;

		ImGui::BeginChild("##pui_res_ubo_table", ImVec2(0, -25));

		// SRVs
		{
			ImGui::Text("SRV");
			ImGui::Separator();
			ImGui::Columns(4);

			// TODO: remove this after imgui fixes the table/column system
			static bool isColumnWidthSet = false;
			if (!isColumnWidthSet) {
				ImGui::SetColumnWidth(0, 80);
				ImGui::SetColumnWidth(1, 60);
				ImGui::SetColumnWidth(2, 80);

				isColumnWidthSet = true;
			}

			ImGui::Text("Controls"); ImGui::NextColumn();
			ImGui::Text("Unit"); ImGui::NextColumn();
			ImGui::Text("Type"); ImGui::NextColumn();
			ImGui::Text("Name"); ImGui::NextColumn();

			ImGui::Separator();

			int id = 0;
			std::vector<GLuint> &els = m_data->Objects.GetBindList(m_modalItem);
			const std::vector<std::string> &items = m_data->Objects.GetObjects();

			/* EXISTING VARIABLES */
			for (const auto& el : els) {
				const std::string& itemName = m_data->Objects.GetItemNameByTextureID(el);

				/* CONTROLS */
				ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
				/* UP BUTTON */
				if (ImGui::Button((UI_ICON_ARROW_UP "##s" + std::to_string(id)).c_str()) && id != 0) {
					GLuint temp = els[id - 1];
					els[id - 1] = el;
					els[id] = temp;
					m_data->Parser.ModifyProject();
				}
				ImGui::SameLine(0,0);
				/* DOWN BUTTON */
				if (ImGui::Button((UI_ICON_ARROW_DOWN "##s" + std::to_string(id)).c_str()) && id != els.size() - 1) {
					GLuint temp = els[id + 1];
					els[id + 1] = el;
					els[id] = temp;
					m_data->Parser.ModifyProject();
				}
				ImGui::PopStyleColor();
				ImGui::NextColumn();

				/* ACTIVE */
				ImGui::Text("%d", id);
				ImGui::NextColumn();

				/* NAME */
				if (objs->IsImage(itemName)) ImGui::Text("image");
				else if (objs->IsImage3D(itemName)) ImGui::Text("image3D");
				else if (objs->IsRenderTexture(itemName)) ImGui::Text("render texture");
				else if (objs->IsAudio(itemName)) ImGui::Text("audio");
				else if (objs->IsBuffer(itemName)) ImGui::Text("buffer");
				else if (objs->IsCubeMap(itemName)) ImGui::Text("cubemap");
				else ImGui::Text("texture");
				ImGui::NextColumn();

				/* VALUE */
				ImGui::Text("%s", itemName.c_str());
				ImGui::NextColumn();

				id++;
			}

			ImGui::Columns(1);
		}

		ImGui::NewLine();

		// UAVs
		{
			ImGui::Text("UAVs / UBOs");
			ImGui::Separator();
			ImGui::Columns(4);

			// TODO: remove this after imgui fixes the table/column system
			static bool isColumnWidthSet = false;
			if (!isColumnWidthSet) {
				ImGui::SetColumnWidth(0, 80);
				ImGui::SetColumnWidth(1, 60);
				ImGui::SetColumnWidth(2, 80);

				isColumnWidthSet = true;
			}

			ImGui::Text("Controls"); ImGui::NextColumn();
			ImGui::Text("Unit"); ImGui::NextColumn();
			ImGui::Text("Type"); ImGui::NextColumn();
			ImGui::Text("Name"); ImGui::NextColumn();

			ImGui::Separator();

			int id = 0;
			std::vector<GLuint> &els = m_data->Objects.GetUniformBindList(m_modalItem);
			const std::vector<std::string> &items = m_data->Objects.GetObjects();

			/* EXISTING VARIABLES */
			for (const auto& el : els) {
				const std::string& itemName = m_data->Objects.GetItemNameByTextureID(el);

				/* CONTROLS */
				ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
				/* UP BUTTON */
				if (ImGui::Button((UI_ICON_ARROW_UP "##u" + std::to_string(id)).c_str()) && id != 0) {
					GLuint temp = els[id - 1];
					els[id - 1] = el;
					els[id] = temp;
					m_data->Parser.ModifyProject();
				}
				ImGui::SameLine(0,0);
				/* DOWN BUTTON */
				if (ImGui::Button((UI_ICON_ARROW_DOWN "##u" + std::to_string(id)).c_str()) && id != els.size() - 1) {
					GLuint temp = els[id + 1];
					els[id + 1] = el;
					els[id] = temp;
					m_data->Parser.ModifyProject();
				}
				ImGui::PopStyleColor();
				ImGui::NextColumn();

				/* ACTIVE */
				ImGui::Text("%d", id);
				ImGui::NextColumn();

				/* NAME */
				if (objs->IsImage(itemName)) ImGui::Text("image");
				else if (objs->IsImage3D(itemName)) ImGui::Text("image3D");
				else if (objs->IsRenderTexture(itemName)) ImGui::Text("render texture");
				else if (objs->IsAudio(itemName)) ImGui::Text("audio");
				else if (objs->IsBuffer(itemName)) ImGui::Text("buffer");
				else if (objs->IsCubeMap(itemName)) ImGui::Text("cubemap");
				else ImGui::Text("texture");
				ImGui::NextColumn();

				/* VALUE */
				ImGui::Text("%s", itemName.c_str());
				ImGui::NextColumn();

				id++;
			}

			ImGui::Columns(1);
		}
		
		ImGui::EndChild();
	}
	void PipelineUI::m_renderMacroManagerUI()
	{
		static ShaderMacro addMacro = { true, "\0", "\0" };
		static bool scrollToBottom = false;

		ImGui::TextWrapped("Add or remove shader macros.");

		ImGui::BeginChild("##pui_macro_table", ImVec2(0, -25));
		ImGui::Columns(4);

		ImGui::Text("Controls"); ImGui::NextColumn();
		ImGui::Text("Active"); ImGui::NextColumn();
		ImGui::Text("Name"); ImGui::NextColumn();
		ImGui::Text("Value"); ImGui::NextColumn();

		ImGui::Separator();

		ImGui::PushStyleColor(ImGuiCol_FrameBg, ImVec4(0, 0, 0, 0));

		int id = 0;
		bool isCompute = m_modalItem->Type == PipelineItem::ItemType::ComputePass;
		bool isAudio = m_modalItem->Type == PipelineItem::ItemType::AudioPass;
		std::vector<ShaderMacro> &els = isCompute ? ((ed::pipe::ComputePass*)m_modalItem->Data)->Macros : (isAudio ? ((ed::pipe::AudioPass*)m_modalItem->Data)->Macros : ((ed::pipe::ShaderPass*)m_modalItem->Data)->Macros);

		/* EXISTING VARIABLES */
		for (auto& el : els) {
			/* CONTROLS */
			ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
			/* UP BUTTON */
			if (ImGui::Button((UI_ICON_ARROW_UP "##" + std::to_string(id)).c_str()) && id != 0) {
				ed::ShaderMacro temp = els[id - 1];
				els[id - 1] = el;
				els[id] = temp;
				m_data->Parser.ModifyProject();
			}
			ImGui::SameLine(0,0);
			/* DOWN BUTTON */
			if (ImGui::Button((UI_ICON_ARROW_DOWN "##" + std::to_string(id)).c_str()) && id != els.size() - 1) {
				ed::ShaderMacro temp = els[id + 1];
				els[id + 1] = el;
				els[id] = temp;
				m_data->Parser.ModifyProject();
			}
			ImGui::SameLine(0,0);
			/* DELETE BUTTON */
			if (ImGui::Button((UI_ICON_DELETE "##" + std::to_string(id)).c_str())) {
				els.erase(els.begin() + id);
				id--;
				m_data->Parser.ModifyProject();
			}
			ImGui::PopStyleColor();
			ImGui::NextColumn();

			/* ACTIVE */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			if (ImGui::Checkbox(("##pui_mcr_act" + std::to_string(id)).c_str(), &el.Active))
				m_data->Parser.ModifyProject();
			ImGui::NextColumn();

			/* NAME */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			if (ImGui::InputText(("##mcrName" + std::to_string(id)).c_str(), el.Name, 32))
				m_data->Parser.ModifyProject();
			ImGui::NextColumn();

			/* VALUE */
			ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
			if (ImGui::InputText(("##mcrVal" + std::to_string(id)).c_str(), el.Value, 512))
				m_data->Parser.ModifyProject();
			ImGui::NextColumn();

			id++;
		}

		ImGui::PopStyleColor();

		// widgets for adding a macro
		/* ADD BUTTON */
		ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		if (ImGui::Button(UI_ICON_ADD)) {
			// cant have two macros with same name
			bool exists = false;
			for (auto& el : els)
				if (strcmp(el.Name, addMacro.Name) == 0)
					exists = true;

			// add if it doesnt exist
			if (!exists) {
				els.push_back(addMacro);
				scrollToBottom = true;
				m_data->Parser.ModifyProject();
			}
		}
		ImGui::NextColumn();
		ImGui::PopStyleColor();
		ImGui::PopItemWidth();

		/* ACTIVE */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		ImGui::Checkbox(("##pui_mcrActAdd" + std::to_string(id)).c_str(), &addMacro.Active);
		ImGui::NextColumn();

		/* NAME */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		ImGui::InputText(("##mcrNameAdd" + std::to_string(id)).c_str(), addMacro.Name, 32);
		ImGui::NextColumn();

		/* VALUE */
		ImGui::PushItemWidth(-ImGui::GetStyle().FramePadding.x);
		ImGui::InputText(("##mcrValAdd" + std::to_string(id)).c_str(), addMacro.Value, 512);
		ImGui::NextColumn();
		
		if (scrollToBottom) {
			ImGui::SetScrollHere();
			scrollToBottom = false;
		}

		ImGui::EndChild();
		ImGui::Columns(1);
	}

	void PipelineUI::m_addShaderPass(ed::PipelineItem *item)
	{
		ed::pipe::ShaderPass *data = (ed::pipe::ShaderPass *)item->Data;

		std::string expandTxt = UI_ICON_UP;
		for (int i = 0; i < m_expandList.size(); i++)
			if (m_expandList[i] == data)
			{
				expandTxt = UI_ICON_DOWN;
				break;
			}

		expandTxt += "##passexp_" + std::string(item->Name);

		ImGui::PushStyleColor(ImGuiCol_Button, ImVec4(0, 0, 0, 0));
		if (ImGui::Button(std::string(std::string(data->Active ? UI_ICON_EYE : UI_ICON_EYE_BLOCKED) + "##hide" + std::string(item->Name)).c_str(), BUTTON_ICON_SIZE))
		{
			data->Active = !data->Active;
			m_data->Parser.ModifyProject();
		}
		ImGui::SameLine(0,0);
		if (ImGui::Button(expandTxt.c_str(), BUTTON_ICON_SIZE))
		{
			if (expandTxt.find(UI_ICON_DOWN) != std::string::npos)
			{
				for (int i = 0; i < m_expandList.size(); i++)
					if (m_expandList[i] == data)
					{
						m_expandList.erase(m_expandList.begin() + i);
						break;
					}
			}
			else
				m_expandList.push_back(data);
		}
		ImGui::PopStyleColor();
		ImGui::SameLine();

		int ewCount = m_data->Messages.GetGroupErrorAndWarningMsgCount(item->Name);
		if (ewCount > 0)
			ImGui::PushStyleColor(ImGuiCol_Text, ThemeContainer::Instance().GetTextEditorStyle(Settings::Instance().Theme)[(int)TextEditor::PaletteIndex::ErrorMessage]);
		
		if (!data->Active)
			ImGui::PushStyleVar(ImGuiStyleVar_Alpha, ImGui::GetStyle().Alpha * 0.5f);
			
		ImGui::Indent(PIPELINE_SHADER_PASS_INDENT);
		if (ImGui::Selectable(item->Name, false, ImGuiSelectableFlags_AllowDoubleClick))
			if (ImGui::IsMouseDoubleClicked(0))
			{
				if (Settings::Instance().General.OpenShadersOnDblClk)
				{
					if (Settings::Instance().General.UseExternalEditor && m_data->Parser.GetOpenedFile() == "")
						m_ui->SaveAsProject(true);
					else
					{
						CodeEditorUI *editor = (reinterpret_cast<CodeEditorUI *>(m_ui->Get(ViewID::Code)));
						if (m_data->Parser.FileExists(data->VSPath))
							editor->OpenVS(item);

						if (m_data->Parser.FileExists(data->PSPath))
							editor->OpenPS(item);

						if (data->GSUsed && strlen(data->GSPath) > 0 && m_data->Parser.FileExists(data->GSPath))
							editor->OpenGS(item);
					}
				}

				if (Settings::Instance().General.ItemPropsOnDblCLk)
				{
					PropertyUI *props = reinterpret_cast<PropertyUI *>(m_ui->Get(ViewID::Properties));
					props->Open(item);
				}
			}

		if (ImGui::BeginDragDropTarget()) {
			if (const ImGuiPayload* payload = ImGui::AcceptDragDropPayload("PipelineItemPayload"))
			{
				// TODO: m_data->Pipeline.DuplicateItem() ?
				ed::PipelineItem* dropItem = *(reinterpret_cast<ed::PipelineItem**>(payload->Data));
				bool duplicate = ImGui::GetIO().KeyCtrl;

				// first find a name that is not used
				std::string name = std::string(dropItem->Name);

				if (duplicate) { // change
					// remove numbers at the end of the string
					size_t lastOfLetter = std::string::npos;
					for (size_t j = name.size() - 1; j > 0; j--)
						if (!std::isdigit(name[j])) {
							lastOfLetter = j + 1;
							break;
						}
					if (lastOfLetter != std::string::npos)
						name = name.substr(0, lastOfLetter);

					// add number to the string and check if it already exists
					for (size_t j = 2; /*WE WILL BRAKE FROM INSIDE ONCE WE FIND THE NAME*/; j++) {
						std::string newName = name + std::to_string(j);
						bool has = m_data->Pipeline.Has(newName.c_str());

						if (!has) {
							name = newName;
							break;
						}
					}
				}

				// get item owner
				void* itemData = nullptr;

				// once we found a name, duplicate the properties:
				// duplicate geometry object:
				if (dropItem->Type == PipelineItem::ItemType::Geometry) {
					pipe::GeometryItem* newData = new pipe::GeometryItem();
					pipe::GeometryItem* origData = (pipe::GeometryItem*)dropItem->Data;
					
					newData->Position = origData->Position;
					newData->Rotation = origData->Rotation;
					newData->Scale = origData->Scale;
					newData->Size = origData->Size;
					newData->Topology = origData->Topology;
					newData->Type = origData->Type;

					if (newData->Type == pipe::GeometryItem::GeometryType::Cube)
						newData->VAO = eng::GeometryFactory::CreateCube(newData->VBO, newData->Size.x, newData->Size.y, newData->Size.z, data->InputLayout);
					else if (newData->Type == pipe::GeometryItem::Circle) {
						newData->VAO = eng::GeometryFactory::CreateCircle(newData->VBO, newData->Size.x, newData->Size.y, data->InputLayout);
						newData->Topology = GL_TRIANGLE_STRIP;
					}
					else if (newData->Type == pipe::GeometryItem::Plane)
						newData->VAO = eng::GeometryFactory::CreatePlane(newData->VBO, newData->Size.x, newData->Size.y, data->InputLayout);
					else if (newData->Type == pipe::GeometryItem::Rectangle)
						newData->VAO = eng::GeometryFactory::CreatePlane(newData->VBO, 1, 1, data->InputLayout);
					else if (newData->Type == pipe::GeometryItem::Sphere)
						newData->VAO = eng::GeometryFactory::CreateSphere(newData->VBO, newData->Size.x, data->InputLayout);
					else if (newData->Type == pipe::GeometryItem::Triangle)
						newData->VAO = eng::GeometryFactory::CreateTriangle(newData->VBO, newData->Size.x, data->InputLayout);
					else if (newData->Type == pipe::GeometryItem::ScreenQuadNDC)
						newData->VAO = eng::GeometryFactory::CreateScreenQuadNDC(newData->VBO, data->InputLayout);
					
					itemData = newData;
				}

				// duplicate Model:
				else if (dropItem->Type == PipelineItem::ItemType::Model) {
					pipe::Model* newData = new pipe::Model();
					pipe::Model* origData = (pipe::Model*)dropItem->Data;

					strcpy(newData->Filename, origData->Filename);
					strcpy(newData->GroupName, origData->GroupName);
					newData->OnlyGroup = origData->OnlyGroup;
					newData->Scale = origData->Scale;
					newData->Position = origData->Position;
					newData->Rotation = origData->Rotation;


					if (strlen(newData->Filename) > 0) {
						std::string objMem = m_data->Parser.LoadProjectFile(newData->Filename);
						eng::Model* mdl = m_data->Parser.LoadModel(newData->Filename);

						bool loaded = mdl != nullptr;
						if (loaded)
							newData->Data = mdl;
						else m_data->Messages.Add(ed::MessageStack::Type::Error, item->Name, "Failed to create .obj model " + std::string(item->Name));
					}

					itemData = newData;
				}

				if (!duplicate) {
					PropertyUI* props = ((PropertyUI*)m_ui->Get(ViewID::Properties));
					if (props->HasItemSelected() && props->CurrentItemName() == dropItem->Name)
						props->Open(nullptr);

					PreviewUI* prev = ((PreviewUI*)m_ui->Get(ViewID::Preview));
					if (prev->IsPicked(dropItem))
						prev->Pick(nullptr);

					m_data->Pipeline.Remove(dropItem->Name);
				}

				m_data->Pipeline.AddItem(item->Name, name.c_str(), dropItem->Type, itemData);
			}
			ImGui::EndDragDropTarget();
		}

		ImGui::Unindent(PIPELINE_SHADER_PASS_INDENT);
		
		if (!data->Active)
			ImGui::PopStyleVar();
			
		if (ewCount > 0)
			ImGui::PopStyleColor();
	}
	void PipelineUI::m_addComputePass(ed::PipelineItem *item)
	{
		ed::pipe::ComputePass *data = (ed::pipe::ComputePass *)item->Data;

		int ewCount = m_data->Messages.GetGroupErrorAndWarningMsgCount(item->Name);
		if (ewCount > 0)
			ImGui::PushStyleColor(ImGuiCol_Text, ThemeContainer::Instance().GetTextEditorStyle(Settings::Instance().Theme)[(int)TextEditor::PaletteIndex::ErrorMessage]);
		else
			ImGui::PushStyleColor(ImGuiCol_Text, ThemeContainer::Instance().GetCustomStyle(Settings::Instance().Theme).ComputePass);

		ImGui::Indent(PIPELINE_SHADER_PASS_INDENT);
		if (ImGui::Selectable(item->Name, false, ImGuiSelectableFlags_AllowDoubleClick))
			if (ImGui::IsMouseDoubleClicked(0))
			{
				if (Settings::Instance().General.OpenShadersOnDblClk)
				{
					if (Settings::Instance().General.UseExternalEditor && m_data->Parser.GetOpenedFile() == "")
						m_ui->SaveAsProject(true);
					else
					{
						CodeEditorUI *editor = (reinterpret_cast<CodeEditorUI *>(m_ui->Get(ViewID::Code)));
						if (m_data->Parser.FileExists(data->Path))
							editor->OpenCS(item);
					}
				}

				if (Settings::Instance().General.ItemPropsOnDblCLk)
				{
					PropertyUI *props = reinterpret_cast<PropertyUI *>(m_ui->Get(ViewID::Properties));
					props->Open(item);
				}
			}
		ImGui::Unindent(PIPELINE_SHADER_PASS_INDENT);
		ImGui::PopStyleColor();
	}
	void PipelineUI::m_addAudioPass(ed::PipelineItem *item)
	{
		ed::pipe::AudioPass *data = (ed::pipe::AudioPass *)item->Data;

		int ewCount = m_data->Messages.GetGroupErrorAndWarningMsgCount(item->Name);
		if (ewCount > 0)
			ImGui::PushStyleColor(ImGuiCol_Text, ThemeContainer::Instance().GetTextEditorStyle(Settings::Instance().Theme)[(int)TextEditor::PaletteIndex::ErrorMessage]);
		else
			ImGui::PushStyleColor(ImGuiCol_Text, ThemeContainer::Instance().GetCustomStyle(Settings::Instance().Theme).ComputePass);

		ImGui::Indent(PIPELINE_SHADER_PASS_INDENT);
		if (ImGui::Selectable(item->Name, false, ImGuiSelectableFlags_AllowDoubleClick))
			if (ImGui::IsMouseDoubleClicked(0))
			{
				if (Settings::Instance().General.OpenShadersOnDblClk)
				{
					if (Settings::Instance().General.UseExternalEditor && m_data->Parser.GetOpenedFile() == "")
						m_ui->SaveAsProject(true);
					else
					{
						CodeEditorUI *editor = (reinterpret_cast<CodeEditorUI *>(m_ui->Get(ViewID::Code)));
						if (m_data->Parser.FileExists(data->Path))
							editor->OpenPS(item);
					}
				}

				if (Settings::Instance().General.ItemPropsOnDblCLk)
				{
					PropertyUI *props = reinterpret_cast<PropertyUI *>(m_ui->Get(ViewID::Properties));
					props->Open(item);
				}
			}
		ImGui::Unindent(PIPELINE_SHADER_PASS_INDENT);
		ImGui::PopStyleColor();
	}
	void PipelineUI::m_addItem(ed::PipelineItem* item)
	{
		ImGui::Indent(PIPELINE_ITEM_INDENT);
		if (ImGui::Selectable(item->Name, false, ImGuiSelectableFlags_AllowDoubleClick))
			if (ImGui::IsMouseDoubleClicked(0)) {
				if (Settings::Instance().General.ItemPropsOnDblCLk) {
					PropertyUI* props = reinterpret_cast<PropertyUI*>(m_ui->Get(ViewID::Properties));
					props->Open(item);
				}

				if (Settings::Instance().General.SelectItemOnDblClk) {
					if (item->Type == PipelineItem::ItemType::Geometry ||
						item->Type == PipelineItem::ItemType::Model)
					{
						bool proceed = true;
						if (item->Type == PipelineItem::ItemType::Geometry)
							proceed = ((pipe::GeometryItem*)item->Data)->Type != pipe::GeometryItem::GeometryType::Rectangle &&
									  ((pipe::GeometryItem*)item->Data)->Type != pipe::GeometryItem::GeometryType::ScreenQuadNDC;
					
						if (proceed) {
							PreviewUI* props = reinterpret_cast<PreviewUI*>(m_ui->Get(ViewID::Preview));
							props->Pick(item);
							m_data->Renderer.AddPickedItem(item);
						}
					}
				}
			}
		if (ImGui::BeginDragDropSource(ImGuiDragDropFlags_None))
		{
			ImGui::SetDragDropPayload("PipelineItemPayload", &item, sizeof(ed::PipelineItem**));
			ImGui::EndDragDropSource();
		}
		ImGui::Unindent(PIPELINE_ITEM_INDENT);
	}
}